#include "qtudptest.h"
#include <qbytearray.h>
#include <qhostaddress.h>
#include <QDebug>
#include <QNetworkInterface>

QtUdpTest::QtUdpTest(QWidget *parent) : QMainWindow(parent) {
	ui.setupUi(this);
	connect(&_Socket, &QUdpSocket::readyRead, this, &QtUdpTest::OnDataReady);
	
	ui.interfaces->addItem("Any");
	ui.interfaces->addItem("Any IPv4");
	ui.interfaces->addItem("Any IPv6");

	QList<QHostAddress> Addresses = QNetworkInterface::allAddresses();
	for (QHostAddress& Address : Addresses) {
		ui.interfaces->addItem(Address.toString());
	}
	
	connect(&_Timer, &QTimer::timeout, this, &QtUdpTest::OnTimeout);
	_Timer.setInterval(1000);
	_Timer.start();
	_ElapsedTimer.start();
}

QtUdpTest::~QtUdpTest() {
}

void QtUdpTest::DoSend() {
	QByteArray Data(ui.msgSize->value(), '\0');

	QHostAddress Address(ui.ipAddress->text());
	quint16 Port = ui.ipPort->value();
	
	unsigned int Flags = 1;

	if (ui.noFragment->isChecked() == false) {
		Flags = 0;
	}

	int ret = ::setsockopt(_Socket.socketDescriptor(), IPPROTO_IP, IP_DONTFRAGMENT, (const char*) &Flags, sizeof(Flags));
	
	*reinterpret_cast<quint64*>(Data.data()) = _ElapsedTimer.nsecsElapsed();

	qint64 r = _Socket.writeDatagram(Data.data(), Data.size(), Address, Port);

	if (r == -1) {
		ui.txtLog->append(QString("Error sending datagram (%1)").arg(_Socket.errorString()));
	} else {
		ui.txtLog->append(QString("%4: Sent %1 bytes to %2:%3...").arg(r).arg(ui.ipAddress->text()).arg(ui.ipPort->value()).arg(QDateTime::currentDateTime().toString("hh:mm:ss,zzz")));
	}

	_Timer.setProperty("LastWriteTime", QDateTime::currentDateTime());
}

void QtUdpTest::OnDataReady() {
	while (_Socket.hasPendingDatagrams()) {
		QByteArray Data(_Socket.pendingDatagramSize(), Qt::Uninitialized);
		
		QHostAddress Address;
		quint16 Port;
		qint64 r = _Socket.readDatagram(Data.data(), Data.size(), &Address, &Port);
		
		if (r == -1) {
			ui.txtLog->append(QString("Error receiving datagram (%1)").arg(_Socket.errorString()));
		} else {
			if (*reinterpret_cast<quint64*>(Data.data() + sizeof(quint64)) != 0) {
				quint64 e = _ElapsedTimer.nsecsElapsed() - *reinterpret_cast<quint64*>(Data.data());
				QString es;
				
				if (e > 1000000) {
					es = QString("%1 ms").arg(e / 1000000.0, 0, 'f', 2);
				} else {
					es = QString("%1 s").arg(e / 1000.0, 0, 'f', 2);
				}

				ui.txtLog->append(QString("%4: Received %1 bytes from %2:%3 (%5 echo)").arg(r).arg(Address.toString()).arg(Port).arg(QDateTime::currentDateTime().toString("hh:mm:ss,zzz")).arg(es));
			} else {
				ui.txtLog->append(QString("%4: Received %1 bytes from %2:%3").arg(r).arg(Address.toString()).arg(Port).arg(QDateTime::currentDateTime().toString("hh:mm:ss,zzz")));
			}

			if (ui.chkEcho->isChecked()) {
				if (*reinterpret_cast<quint64*>(Data.data() + sizeof(quint64)) != 0) {
					*reinterpret_cast<quint64*>(Data.data()) = *reinterpret_cast<quint64*>(Data.data() + sizeof(quint64));
				}
				*reinterpret_cast<quint64*>(Data.data() + sizeof(quint64)) = _ElapsedTimer.nsecsElapsed();
				qint64 r = _Socket.writeDatagram(Data.data(), Data.size(), Address, Port);
			}
		}
	}
}

void QtUdpTest::DoBind() {
	_Socket.close();

	QHostAddress InterfaceAddress;
	
	if (ui.interfaces->currentIndex() == 0) {
		InterfaceAddress = QHostAddress::Any;
	} else if (ui.interfaces->currentIndex() == 1) {
		InterfaceAddress = QHostAddress::AnyIPv4;
	} else if (ui.interfaces->currentIndex() == 2) {
		InterfaceAddress = QHostAddress::AnyIPv6;
	} else {
		InterfaceAddress = QHostAddress(ui.interfaces->currentText());
	}

	_Socket.bind(InterfaceAddress, ui.bindPort->value());
}

void QtUdpTest::OnTimeout() {
	QDateTime LastWriteTime = _Timer.property("LastWriteTime").toDateTime();
	quint64 s = LastWriteTime.secsTo(QDateTime::currentDateTime());
	ui.lblTimer->setText(QString("%1 s").arg(s));
}